"""
Read strucutures from STATE input files
and read results from STATE output files.
"""
import re
import warnings
import numpy as np
from ase import Atoms
from ase.units import Hartree, Bohr

force_unit = Hartree / Bohr


def read_state_input(input_file):
    """ASE module to read STATE input files.

    Args:
        input_file (file-like object or str): Points to the STATE input file

    Returns:
        object: Object based on ASE's atoms module using the read input.
    """
    # USE PARSER to get information
    info = InputParser(input_file)
    info.parse()

    # Initialize ASE atoms info argments
    kwargs = {}

    # Load atomic coordinate (cartesian/crystal) to kwargs
    pos_opt = info.blocks["ATOMIC_COORDINATES"]["opt"]
    if pos_opt in [None, "CART", "CARTESIAN"]:
        # Converted position
        pos = Bohr * info.blocks["ATOMIC_COORDINATES"]["input"]["cps"]
        kwargs["positions"] = pos
    elif pos_opt in ["CRYSTAL", "CRYS"]:
        pos = info.blocks["ATOMIC_COORDINATES"]["input"]["cps"]
        kwargs["scaled_positions"] = pos
    else:
        raise ValueError(f"Invalid atomic coordinate option = {pos_opt}.")

    # Load cell information
    if "CELL" in list(info.variables):
        clist = [float(cnum) for cnum in info.variables["CELL"]]
        # Converted lattice lengths
        lena, lenb, lenc = Bohr * clist[0], Bohr * clist[1], Bohr * clist[2]
        kwargs["cell"] = [lena, lenb, lenc, clist[3], clist[4], clist[5]]
    elif "CELL" in list(info.blocks):
        cell_matrix = info.blocks["CELL"]["input"]
        # Converted cell matrix
        kwargs["cell"] = Bohr * cell_matrix
    else:
        raise ValueError("Cell information not found.")

    # Load Atomic species
    species = info.blocks["ATOMIC_SPECIES"]["input"]["symbol"]
    ityp = info.blocks["ATOMIC_COORDINATES"]["input"]["ityp"]
    symlist = [species[int(i) - 1] for i in ityp]
    kwargs["symbols"] = symlist

    # Initiate ASE Atoms module
    structure = Atoms(**kwargs)
    return structure


def read_state_output(output_file):
    """ASE module to read STATE output files."""

    data = []
    extract = False
    with open(output_file, encoding="utf-8") as file:
        lines = file.readlines()
        if "The calculation has converged" not in str(lines):
            warnings.warn("The calculation has not converged")
        for line in lines:
            if re.search("CONVERGED", line):
                extract = True
                count_empty_line = 0
            if extract:
                data.append(line.split())
                if len(line.split()) == 0:
                    count_empty_line += 1
                if count_empty_line == 2:
                    extract = False
            if extract is False and len(data) != 0:
                energy = Hartree * float(data[2][1])
                #                f_max      = force_unit * float(data[2][2])
                #                f_rms      = force_unit * float(data[2][3])
                force_data = []
                positions = []
                species = []
                for vip_line in data[6:]:
                    if vip_line == []:
                        break
                    if vip_line[0] != "MD:":
                        break
                    species.append(vip_line[2])
                    positions.append(vip_line[3:6])
                    force_data.append(vip_line[6:9])
                positions = Bohr * np.array(positions, dtype=float)
                structure = Atoms(symbols=species, positions=positions)
                forces = force_unit * np.array(force_data, dtype=float)
                data = []
    return structure, energy, forces


class InputParser:
    """Fully parse all input parameters in file."""

    def __init__(self, file_object):
        """Intialize the class."""
        self.file_object = file_object
        self.info = dict()
        # Initialize class attributes
        self.varlines = None
        self.blklines = None
        self.variables = None
        self.blocks = None
        self.subroutine = None

    def parse(self):
        """Launch parsing procedure."""
        self.varblock()
        self.catchvariable()
        self.catchblock()
        self.manager()
        self.errorcheck()
        self.readblock()

    def show(self):
        """Show the variable and block component of the parsed file."""
        print("VARIABLE \n", self.variables)
        print("BLOCK \n", self.blocks)

    def varblock(self):
        """Read and clean input file, and separate variable and block lines."""
        with open(self.file_object, "r", encoding="utf-8") as file:
            lines = file.read().splitlines()

        # Clean comments and trailing spaces
        clean = []
        for line in lines:
            _line = line.strip()
            if not _line.startswith("#"):
                clean.append(line.strip())
        lines = clean.copy()

        # Separate variables and blocks
        for idx, line in enumerate(lines):
            if line.startswith("&"):
                blkstart_idx = idx
                break
        self.varlines = lines[:blkstart_idx]
        self.blklines = lines[blkstart_idx:]

    def catchvariable(self):
        """Create a dictionary from variable lines."""
        self.variables = dict()
        for line in self.varlines:
            parts = line.split()
            varname = parts[0].upper()
            self.variables[varname] = [x.upper() for x in parts[1:]]

    def catchblock(self):
        """Create a dictionary from block lines (not specific)."""
        self.blocks = dict()
        for line in self.blklines:
            linestr = line.upper()

            if linestr.startswith("&") and not linestr.startswith("&END"):
                # Process first line of a block for name
                # and option(if available)
                parts = line.split()
                blkname = parts[0].lstrip("&")
                blkinp, blkopt = [], None
                # For Blocks with additional options
                # (ex. &COORDINATE + CRYSTAL)
                if len(parts) == 2:
                    blkopt = parts[1]

            elif linestr.startswith("&END"):
                # Finalize block dictionary entry
                if blkopt:
                    blkopt = blkopt.upper()
                self.blocks[blkname] = {"opt": blkopt, "input": blkinp}

            else:
                blkinp.append(line)

    def readblock(self):
        """Further process the block dictionary using specific subroutines."""
        # Applies the appopriate read subroutine to each block
        for key in self.blocks:
            if key.upper() in self.subroutine:
                self.subroutine[key.upper()]()

    def manager(self):
        """Define the subroutine that further processes the block dictionary."""

        def atomic_coordinates_read():
            entry = self.blocks["ATOMIC_COORDINATES"]
            lines = entry["input"].copy()
            entry["input"] = {}
            # The following are the EXACT syntax naming in
            _cps = [line.split()[:3] for line in lines]
            entry["input"]["cps"] = np.array(_cps, dtype=float)
            _iwei = [line.split()[3] for line in lines]
            entry["input"]["iwei"] = _iwei
            _imdtyp = [line.split()[4] for line in lines]
            entry["input"]["imdtyp"] = _imdtyp
            _ityp = [line.split()[5] for line in lines]
            entry["input"]["ityp"] = _ityp

        def atomic_species_read():
            entry = self.blocks["ATOMIC_SPECIES"]
            lines = entry["input"].copy()
            entry["input"] = {}
            _symbol, _mass, _file = [], [], []
            for line in lines:
                # For the moment atomic number is not supported
                _symbol.append(line.split()[0])
                _mass.append(line.split()[1])
                _file.append(line.split()[2])

            entry["input"]["symbol"] = _symbol
            entry["input"]["mass"] = _mass
            entry["input"]["file"] = _file

        def cell_read():
            entry = self.blocks["CELL"]
            _vector = [line.split() for line in entry["input"]]
            entry["input"] = np.array(_vector, dtype=float)

        self.subroutine = {
            "ATOMIC_COORDINATES": atomic_coordinates_read,
            "ATOMIC_SPECIES": atomic_species_read,
            "CELL": cell_read,
        }

    def errorcheck(self):
        """Error definition and check of code breaking circumstances."""
        # CELL Block & CELL Variable incomptibility
        if "CELL" in self.variables and "CELL" in self.blocks:
            raise ValueError("Double CELL declaration in block and variable.")

        # ATOMIC_COORDINATES not declared
        if "ATOMIC_COORDINATES" not in list(self.blocks):
            raise ValueError("ATOMIC_COORDINATES is not declared.")



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['state_interface']

package_data = \
{'': ['*']}

install_requires = \
['ase>=3.22,<4.0', 'flake8>=4.0.1,<5.0.0', 'pylint>=2.12.2,<3.0.0']

setup_kwargs = {
    'name': 'state-interface',
    'version': '1.0.0',
    'description': 'ASE STATE interface',
    'long_description': 'None',
    'author': 'Harry Handoko HALIM',
    'author_email': 'harry@cp.prec.eng.osaka-u.ac.jp',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6.2',
}


setup(**setup_kwargs)



"""STATE Calculator
export ASE_STATE_COMMAND="/path/to/STATE < PREFIX.in > PREFIX.out"
Run STATE jobs.
"""

from ase.units import Bohr, Hartree
from ase.calculators.calculator import FileIOCalculator
from ase.calculators.singlepoint import SinglePointDFTCalculator
from ase.data import chemical_symbols
import numpy as np
from .io import read_state_output

ERROR_TEMPLATE = (
    'Property "%s" not available. Please try running STATE\n'
    "first by calling Atoms.get_potential_energy()."
)

WARN_TEMPLATE = (
    'Property "%s" is None. Typically, this is because the '
    "required information has not been printed by STATE "
    'at a "low" verbosity level (the default). '
    'Please try running STATE with "high" verbosity.'
)

force_unit = Hartree / Bohr


class STATE(FileIOCalculator):
    """
    STATE calculator interface class
    """

    implemented_properties = ["energy", "forces"]
    command = "STATE < PREFIX.in > PREFIX.out"
    discard_results_on_any_change = True

    def __init__(self, label="state", atoms=None, **kwargs):
        FileIOCalculator.__init__(self, label, atoms, **kwargs)
        self.calc = None
        self.label = label

    def write_input(self, atoms, properties=None, system_changes=None):
        FileIOCalculator.write_input(self, atoms, properties, system_changes)
        natom = atoms.get_global_number_of_atoms()
        coords = atoms.get_positions() * 1 / Bohr
        species = list(set(atoms.get_chemical_symbols()))
        atomic_nums = atoms.get_atomic_numbers()
        ntyp = len(species)
        cell = atoms.get_cell() * 1 / Bohr
        constraints = atoms.constraints
        with open(self.label + ".in", "w", encoding="utf-8") as file:
            input_data = self.parameters["input_data"]
            print("#\n#\n#", file=file)
            print("NTYP", ntyp, file=file)
            print("NATM", natom, file=file)
            for parameter in input_data.keys():
                write_parameter(parameter, input_json=input_data, file=file)
            print("&CELL", file=file)
            for cell_vector in cell:
                print(
                    f"{cell_vector[0]: >18.12f}",
                    f"{cell_vector[1]: >18.12f}",
                    f"{cell_vector[2]: >18.12f}",
                    file=file,
                )
            print("&END", file=file)
            print("&ATOMIC_SPECIES", file=file)
            for line in input_data["PSEUDOS"]:
                print(*line, file=file)
            print("&END", file=file)
            print("&ATOMIC_COORDINATES", "CARTESIAN", file=file)
            pseudo_idx = {
                lst[0]: i + 1 for i, lst in enumerate(input_data["PSEUDOS"])
            }

            constraints_idx = (
                constraints[0].get_indices() if len(constraints) > 0 else []
            )
            for i, line in enumerate(coords):
                chemical_symbol = chemical_symbols[atomic_nums[i]]
                imdtyp = 0 if i in constraints_idx else 1
                print(
                    f"{line[0]: >18.12f}",
                    f"{line[1]: >18.12f}",
                    f"{line[2]: >18.12f}",
                    1,
                    imdtyp,
                    pseudo_idx[chemical_symbol],
                    file=file,
                )
            print("&END", file=file)
            if "VDW-DF" in input_data:
                print("&VDW-DF", file=file)
                print("QCUT", input_data["VDW-DF"]["QCUT"], file=file)
                print("NQ", input_data["VDW-DF"]["NQ"], file=file)
                print("&END", file=file)

    def read_results(self):
        structure, energy, forces = read_state_output(self.label + ".out")
        calc = SinglePointDFTCalculator(
            structure, energy=energy, forces=forces
        )

        self.calc = calc
        self.results = calc.results
        self.results["updt_pos"] = structure.get_positions()

    def get_stress(self, atoms):
        """Dummy class method to get stress"""
        return np.zeros((3, 3))

    def get_updt_pos(self):
        """Return final atoms positions"""
        return self.results["updt_pos"]


def write_parameter(parameter, input_json, file):
    """Helper method to write input parameters"""
    if parameter == "KPOINT_MESH":
        print("KPOINT_MESH", *input_json["KPOINT_MESH"], file=file)
    elif parameter == "KPOINT_SHIFT":
        print("KPOINT_SHIFT", *input_json["KPOINT_SHIFT"], file=file)
    elif parameter in ["PSEUDOS", "SUBSTRATE_IDX", "VDW-DF"]:
        pass
    else:
        print(parameter, input_json[parameter], file=file)
